'use strict';

const colors = require('./colors');
const symbols = require('./symbols');
const fs = require('fs');
const stripAnsi = require('strip-ansi');
const path = require('path');
const isInteractiveTerminal = require('is-interactive');
const { PassThrough } = require('stream');

class Output {
  /** @type {NodeJS.WritableStream} */
  stdout;
  /** @type {undefined | NodeJS.WriteStream} Undefined if not interactive */
  interactiveStdout;
  /** @type {NodeJS.WritableStream} */
  stderr;
  /** @type {undefined | NodeJS.WriteStream} Undefined if not interactive */
  interactiveStderr;
  /** @type {undefined | NodeJS.ReadStream} Undefined if not interactive */
  interactiveStdin;
  /** @type {NodeJS.WritableStream} */
  logsFileStream;

  /** @type {Array<{ namespace?: string[], message: string }>} */
  verboseLogs = [];
  logsFilePath = '.serverless/compose.log';

  /**
   * @param {boolean} verboseMode
   * @param {boolean} [disableIO] To allow mocking in tests
   */
  constructor(verboseMode, disableIO = false) {
    this.verboseMode = verboseMode;
    this.stdout = disableIO ? new PassThrough() : process.stdout;
    this.stderr = disableIO ? new PassThrough() : process.stderr;
    this.logsFileStream = disableIO ? new PassThrough() : this.openLogsFile();
    if (!disableIO && isInteractiveTerminal()) {
      this.interactiveStdout = process.stdout;
      this.interactiveStderr = process.stderr;
      this.interactiveStdin = process.stdin;
    }
  }

  /**
   * Writes text to stdout.
   * @param {string} message
   * @param {string[]} [namespace]
   */
  writeText(message, namespace = []) {
    message = this.namespaceLogMessage(message, namespace);
    this.safeWrite(message);
    this.writeToLogsFile(message);
  }

  /**
   * Writes logs to stderr.
   * @param {string} [message]
   * @param {string[]} [namespace]
   */
  log(message, namespace = []) {
    message = this.namespaceLogMessage(message, namespace);
    this.safeWrite(message, false);
    this.writeToLogsFile(message);
  }

  /**
   * @param {string} message
   * @param {string[]} [namespace]
   */
  verbose(message, namespace) {
    if (!message || message === '') return;

    this.writeToLogsFile(this.namespaceLogMessage(message, namespace));

    if (this.verboseMode) {
      this.doLogVerbose(message, namespace);
    } else {
      this.verboseLogs.push({
        message,
        namespace,
      });
    }
  }

  /**
   * @param {string|Error} error
   * @param {string[]} [namespace]
   */
  error(error, namespace = []) {
    if (this.verboseMode && error instanceof Error) {
      // Print the stack trace in verbose mode
      this.log(`${colors.red('Error:')} ${error.stack}`, namespace);
    } else {
      error = error instanceof Error ? error.message : error;
      this.log(`${colors.red('Error:')} ${error}`, namespace);
    }
  }

  enableVerbose() {
    this.verboseMode = true;
    // Flush all previous verbose logs to the output
    this.verboseLogs.forEach(({ message, namespace }) => {
      this.doLogVerbose(message, namespace);
    });
    this.verboseLogs = [];
  }

  /**
   * @private
   * @param {string} message
   * @param {string[]} [namespace]
   */
  doLogVerbose(message, namespace) {
    message = colors.gray(message);
    message = this.namespaceLogMessage(message, namespace);
    this.safeWrite(message, false);
  }

  /**
   * @private
   * @param {string} text
   */
  writeToLogsFile(text) {
    text = stripAnsi(text);
    this.logsFileStream.write(`${text}\n`);
  }

  /**
   * @private
   * @param {string|undefined} text
   * @param {string[]} [namespace]
   * @return {string}
   */
  namespaceLogMessage(text, namespace) {
    text = text ?? '';
    const prefix = this.generatePrefix(namespace);
    return text
      .split('\n')
      .map((line) => `${prefix}${line}`)
      .join('\n');
  }

  /**
   * @private
   * @param {string[]} [namespace]
   * @return string
   */
  generatePrefix(namespace) {
    if (!namespace) {
      return '';
    }
    let prefix = namespace.join(` ${symbols.separator} `);
    if (prefix.length > 0) {
      prefix = colors.gray(`${prefix} ${symbols.separator} `);
    }
    return prefix;
  }

  openLogsFile() {
    const directory = path.dirname(this.logsFilePath);
    if (!fs.existsSync(directory)) {
      fs.mkdirSync(directory);
    }
    return fs.createWriteStream(this.logsFilePath);
  }

  /**
   * Safely writes text to a stream.
   * It takes into account dynamic content that may be written below.
   * @private
   * @param {string | undefined} text
   * @param {boolean} [stdout] If false, stderr is used.
   */
  safeWrite(text, stdout = true) {
    const stream = stdout ? this.stdout : this.stderr;
    const interactiveStream = stdout ? this.interactiveStdout : this.interactiveStderr;

    text = text ?? '';
    for (const line of text.split('\n')) {
      // This writes from the cursor
      stream.write(line);
      // But maybe the line already contained content (e.g. a progress)
      // so we clear the rest of line, up till its end (on the right side)
      if (interactiveStream) {
        interactiveStream.clearLine(1);
      }
      // Then we can add a line return
      stream.write('\n');
    }
  }
}

module.exports = Output;
